#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fr_recon.h"
#include "fr_paramest.h"
#include "fr_icm.h"
#include "../ansi_lib_loc/allocate.h"
#include "../ansi_lib_loc/assort.h"
#include "../ansi_lib_loc/multigrid.h"
#include "../ansi_lib_loc/solve2.h"
#include "../ansi_lib_loc/randlib.h"
#include "../ansi_lib_loc/A_comp.h"
#ifdef __linux__
#include <getopt.h>
#endif
extern int optind;
extern char *optarg;

/*#define RECON_MASK*/

void cmdline_error(void);
void read_data(struct all_data *data,struct column *sp_col,struct cmdline_params *cmdline);
void write_output(struct cmdline_params *cmdline,struct all_data *data,double **sigma,struct geom_info *geom);
void free_memory(struct all_data *data, struct column *sp_col,double **sigma,struct cmdline_params *cmdline);
void init_sp_col(struct column *sp_col,struct sino_info *sino);
double recon(struct all_data *data,struct geom_info *geom,struct column *sp_col,struct cmdline_params *cmdline);
void calculate_pix_recon_mask(struct all_data *data,struct geom_info *geom);
void clip_initial(struct all_data *data,struct geom_info *geom);
void get_cmdline_params(int argc, char **argv, struct cmdline_params *cmdline);
void read_init_and_mask(struct all_data *data,struct cmdline_params *cmdline);




int main( int argc, char **argv )
{
  int N,M;
  struct geom_info geom;
  struct column sp_col;
  struct cmdline_params cmdline;
  struct all_data data;
  double **sigma;

  /* parse command-line */
  get_cmdline_params(argc, argv, &cmdline);


  /* init random generator */
  srandom2(1);
  

  /* read input */
  read_data(&data,&sp_col,&cmdline);

  /* Set basic reconstruction geometry */
  if (cmdline.xdFLAG==0)
    cmdline.xdim=data.sino.n_t;
  if (cmdline.ydFLAG==0)
    cmdline.ydim=data.sino.n_t;
  if (cmdline.fovx_FLAG==0)
    cmdline.fov_x_mm = data.sino.delta_t * (double)data.sino.n_t;

  geom.n_y   = cmdline.ydim;
  geom.n_x   = cmdline.xdim;
  geom.delta = cmdline.fov_x_mm/(double)cmdline.xdim;
  geom.x_0   = -(cmdline.xdim-1)*geom.delta/2.0;
  geom.y_0   = -(cmdline.ydim-1)*geom.delta/2.0;
  N=geom.n_y;
  M=geom.n_x;

  printf("\nReconstruction geometry:\n");
  printf("xdim=%d pixels\nydim=%d pixels\nfield_of_view_x=%f mm\n\n",cmdline.xdim,cmdline.ydim,cmdline.fov_x_mm);
  printf("pixel size=%f mm\n",geom.delta);

  
  /* calc_pix_recon_mask: if pix outside t_0 or outside fov,
     will set to zero */
  calculate_pix_recon_mask(&data,&geom);

  /* Read initialization and mask-file (mask if parameter-estimation is on)*/
  read_init_and_mask(&data,&cmdline);

  clip_initial(&data,&geom);
 

  /* Allocate memory and initialize projection matrices*/
  init_sp_col(&sp_col,&data.sino);
  sigma = (double **)get_img(1,1,sizeof(double));
 
  /* fix-res reconstruction */
  sigma[0][0] = recon(&data,&geom,&sp_col,&cmdline);
  
  write_output(&cmdline,&data,sigma,&geom);

  free_memory(&data, &sp_col,sigma,&cmdline);

  return(0);

}




double recon(struct all_data *data,struct geom_info *geom,struct column *sp_col,struct cmdline_params *cmdline)
{
   
  double **profiles;
  int i;
  static double *S;
  static int param, N,M;
  
  static double SX=0.0,P;

  srandom2(1);

  P = cmdline->p-1;
  
  /* compute pixel profiles, data needed for computing A */
  fflush(stderr);
  printf("Calculating pixel profiles for system matrix...\n");
  fflush(stdout);
  pix_prof_comp(&data->sino,geom,&profiles);
  printf("done.\n\n");fflush(stdout);fflush(stderr);
 
  N = geom->n_y;
  M = geom->n_x;
 
 
  
  /* PARAMETER ESTIMATION */
  if( cmdline->no_iter_param != 0)
    {
      /* calculate number of pixels w/ nonzero mask  Nsup */
      data->Nsup=0;
      for(i=0;i<N*M;i++)
	if(data->mask[0][i])
	  data->Nsup++;
      
      SX=parameter_estimation(data, geom,profiles, cmdline ,sp_col, P);
    }
  else
    SX = pow(cmdline->sigma,cmdline->p);
  
  printf("Sigma is %f\n",pow(SX,1.0/cmdline->p));


  /* reconstruct */
  if(cmdline->no_iter_rec != 0)
    {
      /* Allocate S  which is Ax*/
      S = (double *) get_spc(data->sino.n_t*data->sino.n_theta, sizeof(double));

      /* Calculate forward projection of initial image X, store in S */
      f_pro(S,data,geom,profiles,sp_col);


      printf("Starting reconstruction ...\n");
      param = 0;
      SX=relax_recon(data,S, geom, profiles, cmdline,sp_col,P,SX);


      printf("Finished reconstruction\n");
      free((void *)S);
    }
  
  
  free_img((void **)profiles);
  
  return(pow(SX,1.0/(P+1)));
  
}


void init_sp_col(struct column *sp_col,struct sino_info *sino)
{
  sp_col->val = (double *) get_spc(sino->n_t*sino->n_theta, sizeof(double));
  sp_col->row = (int *) get_spc(sino->n_t*sino->n_theta, sizeof(int));
  sp_col->multiplier = 1.0; 
}





void cmdline_error()
{
  fflush(stdout);fflush(stderr);
  fprintf(stderr,"Usage: trans [-x {#pix_x} (n_t)] [-y {#pix_y} (n_t)] \n     [-f {field_of_view_x_in_mm} (n_t*delta_t)] [-i {#iter_recon} (20)] \n     [-E {#iter_param} (0)] [-s {sigma} <not with -E>] \n     [-p {GGMRF_shape_param} (2.0)]  [-I init_file.mat] [-S support_file] \n     [-R pixelreplicate_result]  inputfile outputfile\n\n\a");
  fprintf(stderr,"Values in () indicate default settings.\n\n");
  fflush(stdout);fflush(stderr);
  exit(-1);
}



void get_cmdline_params(int argc, char **argv, struct cmdline_params *cmdline)
{

  int i,ct;

  cmdline->fovx_FLAG=0;
  cmdline->xdFLAG=0;
  cmdline->ydFLAG=0;
  cmdline->p=2.0;
  cmdline->no_iter_param=0;
  cmdline->no_iter_rec=20;
  cmdline->init_FLAG=0;  
  cmdline->sup_FLAG=0;
  cmdline->sigma_FLAG=0;
  cmdline->interpolate_FLAG=0;
 

  /* get options */
  for (;;)
    { 
      i= getopt(argc, argv, "x:y:f:p:i:E:I:S:s:R");
      if (i==EOF)
	break;
      else
	{
	  if (i=='x')
	    {
	      cmdline->xdim=atoi(optarg);
	      cmdline->xdFLAG=1;
	      }
	  if (i=='y')
	    {
	      cmdline->ydim=atoi(optarg);
	      cmdline->ydFLAG=1;
	    }
	  if (i=='f')
	    {
	      cmdline->fov_x_mm=atof(optarg);
	      cmdline->fovx_FLAG=1;
	    }
	  if (i=='p')
	    {
	      cmdline->p=atof(optarg);
	    }
	  if (i=='i') 
	    {
	      cmdline->no_iter_rec=atoi(optarg);
	    }
	  
	  if (i=='E') 
	    {
	      cmdline->no_iter_param=atoi(optarg);
	    }
	  if (i=='I') 
	    {
	      cmdline->init=optarg;
	      cmdline->init_FLAG=1;
	    }
	  if (i=='S') 
	    {
	      cmdline->support=optarg;
	      cmdline->sup_FLAG=1;
	    }
	  if (i=='s') 
	    {
	      cmdline->sigma=atof(optarg);
	      cmdline->sigma_FLAG=1;
	    }
	  if (i=='R') 
	    cmdline->interpolate_FLAG=1;
	}
    }

  /* Read required arguments */
  if (argc-optind != 2)
    {
      printf("Incorrect no. of arguments\n");
      cmdline_error();
    }
  ct=optind;

  cmdline->input   = argv[ct];
  ct++;
  cmdline->output   = argv[ct];
  ct++;


  /* Check options */
  if (cmdline->no_iter_param!=0) 
    {
      if (cmdline->sigma_FLAG==1)
	{
	  fprintf(stderr,"Error: Can't manually specify sigma and set number of parameter_estimation_iterations>0\n\n");
	  cmdline_error();
	}
      if (cmdline->sup_FLAG==0)
	{
	  fprintf(stderr,"Error: Must specify support file if number of parameter_estimation_iterations>0\n\n");	 
	  cmdline_error();
	}
    }
  else
    {
      if  (cmdline->sigma_FLAG==0)
	{
	  fprintf(stderr,"Error: Must either manually specify sigma OR  set number of parameter_estimation_iterations>0\n\n");
	  cmdline_error();
	}
      if (cmdline->sup_FLAG==1)
	{
	  fprintf(stderr,"NOTE: Support file specified but #iter_param set to zero.\nSupport file will not be needed/used and will be ignored.\n\a\n");
	}
    }


      printf("input file=%s\noutput file=%s\n",cmdline->input,cmdline->output);
      if (cmdline->init_FLAG)
	printf("init file=%s\n",cmdline->init);
      else
	printf("NO init file specified\n");
      if (cmdline->sup_FLAG)
	printf("suppport file=%s\n",cmdline->support);
      else
	printf("NO support file specified\n");
      
      
      printf("reconstruction_iterations=%d\n",cmdline->no_iter_rec);
      if (cmdline->no_iter_param!=0) 
	printf("parameter_estimation_iterations=%d\n",cmdline->no_iter_param);
      else
	{
	  printf("parameter estimation OFF\n");
	  printf("sigma=%f\n",cmdline->sigma);
	}
      printf("GGMRF shape parameter p=%f\n",cmdline->p);
      if (cmdline->interpolate_FLAG)
	printf("Pixel-replication of output reconstruction=ON\n");
      else
	printf("Pixel-replication of output reconstruction=OFF\n");
    
}














/*void cmdline_error()
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"emisnew [-i #iter (10)] [-s sigma (10.0)] [-p GGMRF_param (2.0)] [-f field_of_view_mm (200.0)] [-r recon_resol (128)] [-I interpolate_output]  input.mat initial.mat output.mat\n");
  fprintf(stderr,"\nUse 'none' for constant initial condition.\n\n");
  exit(-1); 
}




void parse_cmdline(struct cmdline_params *cmdline,int argc,char **argv)
{
  int ct=0;
  int i;

  cmdline->no_iter_param=0;
  cmdline->p=2.0;
  cmdline->no_iter_rec=10;
  cmdline->sigma=10.0;
  cmdline->interpolate=0;
  cmdline->fov=200.0;
  cmdline->resol=128;
  cmdline->mask_FLAG=0;

  for (;;)
    { 
      i= getopt(argc, argv, "i:s:p:If:r:E:m:");
      if (i==EOF)
        break;
      else
        {
          if (i == '?')
            cmdline_error();
	  if (i=='i')
	    {
	      cmdline->no_iter_rec=atoi(optarg);
	    }
	  if (i=='s')
	    {
	      cmdline->sigma=atof(optarg);
	    }
	  if (i=='p')
	    {
	      cmdline->p=atof(optarg);
	    }  
	  if (i=='I')
	    {
	      cmdline->interpolate=1;
	    }
	  if (i=='f')
	    {
	      cmdline->fov=atof(optarg);
	    }
	  if (i=='r')
	    {
	      cmdline->resol=atoi(optarg);
	    }
	  if (i=='E')
	    {
	      cmdline->no_iter_param=atoi(optarg);
	    }
	  if (i=='m')
	    {
	      strcpy(cmdline->mask_fname,optarg);  
	      cmdline->mask_FLAG=1;
	    }
	  
	}
    }

  ct=optind;
  
  if ((argc-optind < 3) || (argc-optind > 3))
    {
      cmdline_error();
    }

  cmdline->input   = argv[ct];
  ct++;
  cmdline->initial = argv[ct];
  ct++;
  cmdline->output = argv[ct];
  ct++;


  if (( cmdline->no_iter_param>0) && (cmdline->mask_FLAG==0))
    {
      fprintf(stderr,"Input Error: parameter estimation swicted on (-E flag) but no mask specified (-m).\n\a\n");
      exit(-1);
    }


  printf("\nCommand-line arguments:\n");
  printf(" Input file : %s\n Initial    : %s\n Output file: %s\n",cmdline->input,cmdline->initial,cmdline->output);
  printf(" Parameters : p=%f, sigma=%f, no_iter_rec %d\n",cmdline->p,cmdline->sigma,cmdline->no_iter_rec);
  printf("              fov=%fmm  resol=%dpix\n",cmdline->fov,cmdline->resol);


  printf("\n\n");
}
*/



void read_data(struct all_data *data,struct column *sp_col,struct cmdline_params *cmdline)
{
  

  int row,col;

  FILE *fin;
  
  int n_t,n_theta;
  double delta_theta,delta_t,theta_0,t_0;

  /* read input sinogram Y */
  printf("Reading input sinogram, parameters:\n");
  fflush(stdout);fflush(stderr);
  fin = fopen(cmdline->input, "r");
  if (fin == NULL)
    {
      fprintf(stderr,"Cannot read input file %s\n",cmdline->input);
      exit(-1);
    }
  

  /* read header with sinogram size and parameters */
  fscanf(fin,"%d %d %lf %lf %lf %lf",&n_theta,&n_t,&delta_theta,&delta_t,&theta_0,&t_0);

  data->sino.n_theta=n_theta;
  data->sino.n_t=n_t;
  data->sino.delta_theta=delta_theta;
  data->sino.delta_t=delta_t;
  data->sino.theta_0=theta_0;
  data->sino.t_0=t_0;



  printf("n_theta=%d\nn_t=%d\ndelta_theta=%f\ndelta_t=%f\ntheta_0=%f\nt_0=%f\n\n",n_theta,n_t,delta_theta,delta_t,theta_0,t_0);



  /* read dosage 2-D array */
  data->Yt = (double **)get_img(n_t,n_theta,sizeof(double));
  for (row=0;row<n_theta;row++)
    for (col=0;col<n_t;col++)
      {
	fscanf(fin,"%lg",&data->Yt[row][col]);
      }

  /* read sinogram 2-D array */

  data->Y = (double **)get_img(n_t,n_theta,sizeof(double));
  for (row=0;row<n_theta;row++)
    for (col=0;col<n_t;col++)
      {
	fscanf(fin,"%lg",&data->Y[row][col]);
      }

  fclose(fin);
  fflush(stdout);fflush(stderr);
  


  /*fin=fopen("Y","w");
    for (row=0;row<n_theta;row++)
    {
    for (col=0;col<n_t;col++)
    fprintf(fin,"%g ",data->Y[row][col]);
    fprintf(fin,"\n");
    }
    fclose(fin);
  */


 
  
  printf("Done reading data.\n\n");
  
}




void read_init_and_mask(struct all_data *data,struct cmdline_params *cmdline)
{
  int row,col;
  FILE *fin;

  data->X = (double**)get_img(cmdline->xdim,cmdline->ydim,sizeof(double));
  
  if (cmdline->init_FLAG)
    {
      fin = fopen(cmdline->init, "r");
      if (fin == NULL)
	{
	  fprintf(stderr,"Cannot read file %s\n\a\n",cmdline->init);
	  exit(-1);
	}
      
      for (row=0;row<cmdline->ydim;row++)
	for (col=0;col<cmdline->xdim;col++)
	  fscanf(fin,"%lg",&data->X[row][col]);
      
      fclose(fin);
    }
  else
    {
      for (row=0;row<cmdline->ydim;row++)
	for (col=0;col<cmdline->xdim;col++)
	  if (data->recon_mask[row][col])
	    data->X[row][col]=0.05;
	  else
	    data->X[row][col]=0.0;
    }


  
  /* Mask is only needed when automatic estimation of the
     smoothing parameter is performed */
  if (cmdline->sup_FLAG)
    {
      /* read support = MASK for parameter estimation */
      fin = fopen(cmdline->support, "r");
      if (fin == NULL)
	{ 
	  fprintf(stderr,"Cannot read file %s\n",cmdline->support);
	  exit(-1);
	} 
      
      data->mask = (int **)get_img(cmdline->xdim,cmdline->ydim,sizeof(int)); 
      for (row=0;row<cmdline->ydim;row++)
	for (col=0;col<cmdline->xdim;col++)
	  {
	    fscanf(fin,"%d",&data->mask[row][col]);
	  }

      fclose(fin);
    }
  else
    data->mask=NULL;
}











void write_output(struct cmdline_params *cmdline,struct all_data *data,double **sigma,struct geom_info *geom)
{
  FILE *fout;
  int N,M,i,j;

  double **tmp;
  N=geom->n_y;
  M=geom->n_x;
  printf("Writing output image %s ...\n",cmdline->output);
  
  
  fout = fopen(cmdline->output, "w");
  if (fout == NULL)
    {
      fprintf(stderr,"Cannot write file %s\n",cmdline->output);
      exit(-1);
    }
      

  if (cmdline->interpolate_FLAG==0)
    {
      for (i=0;i<N;i++)
	{
	  for (j=0;j<M;j++) 
	    fprintf(fout,"%g ",data->X[i][j]);
	  fprintf(fout,"\n");
	}
    }
  else
    {
      tmp=(double**)multialloc(sizeof(double),2,2*N,2*M);
      for (i=0;i<2*N;i++)
	for (j=0;j<2*M;j++)
	  tmp[i][j] = data->X[i/2][j/2];
      for (i=0;i<2*N;i++)
	{
	  for (j=0;j<2*M;j++) 
	    fprintf(fout,"%g ",tmp[i][j]);
	  fprintf(fout,"\n");
	}
      multifree((void**)tmp,2);
    }

  /*sprintf(matrix,"sigma");
    write_mat_double_image(sigma,1,1,matrix,fout);
    
    sprintf(matrix,"fov");
    fov=geom->dia/10.0;
    ptmp=&fov;
    write_mat_double_image(&ptmp,1,1,matrix,fout);
  */
  fclose(fout);
}
  


void init_X_to_constant(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col)
{
  double init_const=0.1;
  int i,j;

  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      if (data->recon_mask[i][j])
	data->X[i][j]=init_const;
      else
	data->X[i][j]=0.0;
}





void calculate_pix_recon_mask(struct all_data *data,struct geom_info *geom)
{
  int i,j;
  double r_sq,x,y,r_sq_max;
  
  r_sq_max = data->sino.t_0 * data->sino.t_0 ;
  x=y=r_sq=r_sq_max=0.0;


  data->recon_mask=(int**)multialloc(sizeof(int),2,geom->n_y,geom->n_x);
  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      data->recon_mask[i][j]=1.0;

#ifdef RECON_MASK
  r_sq_max = data->sino.t_0 * data->sino.t_0 ;
  


  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      {
	y = geom->y_0 + i*geom->delta;
	x = geom->x_0 + j*geom->delta;
	
	r_sq=x*x+y*y;
	
	if (r_sq>r_sq_max)
	  data->recon_mask[i][j]=0.0;
      }
#endif
}
  

void clip_initial(struct all_data *data,struct geom_info *geom)
{
  int i,j;
  
  for (i=0;i<geom->n_y;i++)
    for (j=0;j<geom->n_x;j++)
      if ((data->recon_mask[i][j]==0) || (data->X[i][j]<0.0))
	data->X[i][j]=0.0;
}


 
void free_memory(struct all_data *data, struct column *sp_col,double **sigma,struct cmdline_params *cmdline)
{
  printf("Freeing memory ...\n");
  
  free_img((void **)data->X);
  free_img((void **)sigma);
  free_img((void **)data->Y);
  free_img((void **)data->Yt);
  free((void *)sp_col->val);
  free((void *)sp_col->row);
  free_img((void **)data->recon_mask);

  if (cmdline->sup_FLAG)
    free_img((void **)data->mask);

}
